package controlisolation.lifecyclefsm.stateexecutor;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import controlisolation.lifecyclefsm.LifeCycleFSM;

public class StateExecutor {

    private String _name;
    private StateWork _work;
    private List _workStartedListeners, _workFinishedListeners, _workFailedListeners,
            _workWarningsListeners;
    private static final Logger _logger =
            Logger.getLogger(StateExecutor.class.getName());
    private LifeCycleFSM _containerFSM;

    public StateExecutor(String name, LifeCycleFSM fsm) {
        _workStartedListeners = new ArrayList();
        _workFinishedListeners = new ArrayList();
        _workFailedListeners = new ArrayList();
        _workWarningsListeners = new ArrayList();
        _name = name;
        _containerFSM = fsm;

    }

    public void doWork() {
        if (_work == null) {
            //_logger.log(Level.INFO, "No work defined for state:" + _name);
            return;
        }


        if (_containerFSM.isDebugMessages()) {
            _logger.log(Level.INFO, "STARTING: " + _name);
        }
        _work.getWorkOutputs().clear();

        if (_containerFSM.isDebugMessages()) {
            logStateInputInfo();
        }
        fireWorkStartedEvent();
       

        try {
            _work.doWork();
        } catch (Exception e) {
            _work.addException(e);
        }

        if (_containerFSM.isDebugMessages()) {
            logStateOutputInfo();
        }
        fireWorkFinishedEvent();
        if (!_work.suscessful()) {
            if (_containerFSM.isDebugMessages()) {
                logWorkErrors();
            }
            fireWorkFailedEvent();
        } else if (_work.hasWarnings()) {
            if (_containerFSM.isDebugMessages()) {
                logWorkWarnings();
            }
            fireWorkProducedWarningsEvent();
        }

        if (_containerFSM.isDebugMessages()) {
            _logger.log(Level.INFO, "FINISHED: " + _name);
        }
    }

    /**
     * @return the _name
     */
    public String getName() {
        return _name;
    }

    /**
     * @param _name the _name to set
     */
    public void setName(String name) {
        this._name = name;
    }

    /**
     * @return the _work
     */
    public StateWork getWork() {
        return _work;
    }

    /**
     * @param work the _work to set
     */
    public void setWork(StateWork work) {
        this._work = work;
        work.setStateExecutor(this);
    }

    public synchronized void addWorkFailedEventListener(WorkFailedEventListener listener) {
        _workFailedListeners.add(listener);
    }

    public synchronized void removeWorkFailedEventListener(WorkStartedEventListener listener) {
        _workFailedListeners.remove(listener);
    }

    public synchronized void addWorkStartedEventListener(WorkStartedEventListener listener) {
        _workStartedListeners.add(listener);
    }

    public synchronized void removeWorkStartedEventListener(WorkStartedEventListener listener) {
        _workStartedListeners.remove(listener);
    }

    public synchronized void addWorkFinishedEventListener(WorkFinishedEventListener listener) {
        _workFinishedListeners.add(listener);
    }

    public synchronized void removeWorkFinishedEventListener(WorkFinishedEventListener listener) {
        _workFinishedListeners.remove(listener);
    }

    private synchronized void fireWorkStartedEvent() {
        WorkStartedEvent event = new WorkStartedEvent(this._work);
        Iterator i = _workStartedListeners.iterator();
        while (i.hasNext()) {
            ((WorkStartedEventListener) i.next()).handleWorkStartedEvent(event);
        }
    }

    private synchronized void fireWorkFinishedEvent() {
        WorkFinishedEvent event = new WorkFinishedEvent(this._work);
        Iterator i = _workFinishedListeners.iterator();
        while (i.hasNext()) {
            ((WorkFinishedEventListener) i.next()).handleWorkFinishedEvent(event);
        }
    }

    private synchronized void fireWorkFailedEvent() {
        WorkFailedEvent event = new WorkFailedEvent(this._work);
        Iterator i = _workFailedListeners.iterator();
        while (i.hasNext()) {
            ((WorkFailedEventListener) i.next()).handleWorkFailedEvent(event);
        }
    }

    private synchronized void fireWorkProducedWarningsEvent() {
        WorkWarningsEvent event = new WorkWarningsEvent(this._work);
        Iterator i = this._workWarningsListeners.iterator();
        while (i.hasNext()) {
            ((WorkWarningsEventListener) i.next()).handleWarningsProducedEvent(event);
        }
    }

    public void addWorkInput(String name, Object value) {
        this._work.getWorkInputs().put(name, value);
    }

    public void setWorkInputs(Map<String, Object> work_inputs) {
        this._work.setWorkInputs(work_inputs);
    }

    public Map<String, Object> getWorkOutputs() {
        return this._work.getWorkOutputs();
    }

    public List<String> getWorkErrors() {
        return this._work.getWorkErrors();
    }

    private void logWorkErrors() {
        if (!this._work.getWorkExceptions().isEmpty()) {
            _logger.log(Level.SEVERE, "WORK FATAL EXCEPTION:");
            for (Exception e : _work.getWorkExceptions()) {
                _logger.log(Level.SEVERE, "\t" + e.toString());
            }

        }

        if (!_work.getWorkErrors().isEmpty()) {
            _logger.log(Level.SEVERE, "WORK ERRORS:");
        }
        for (String str : _work.getWorkErrors()) {
            _logger.log(Level.SEVERE, "\t" + str);
        }



    }

    private void logWorkWarnings() {

        if (this._work.getWorkWarnings() != null) {
            _logger.log(Level.WARNING, "WORK WARNING MESSAGES:");
        }

        for (String str : _work.getWorkWarnings()) {
            _logger.log(Level.WARNING, "\t" + str);
        }

    }

    private void logStateInputInfo() {

        _logger.log(Level.INFO, StateWork.toStringWorkInputs(this.getWork()));
    }

    private void logStateOutputInfo() {

        _logger.log(Level.INFO, StateWork.toStringWorkOutputs(this.getWork()));
    }

    public LifeCycleFSM getContainerFSM() {
        return _containerFSM;
    }
}
